#!/usr/bin/env python3

import os,sys,struct

import u3 as u3py

class ClockStuckException(Exception): pass
class NackException(Exception): pass

# error codes from
# https://labjack.com/support/datasheets/u3/low-level-function-reference/errorcodes
# example code uses magic numbers(!)

SCRATCH_WRT_FAIL = 1
SCRATCH_ERASE_FAIL = 2
DATA_BUFFER_OVERFLOW = 3
ADC0_BUFFER_OVERFLOW = 4
FUNCTION_INVALID = 5
SWDT_TIME_INVALID = 6
XBR_CONFIG_ERROR = 7

FLASH_WRITE_FAIL = 16
FLASH_ERASE_FAIL = 17
FLASH_JMP_FAIL = 18
FLASH_PSP_TIMEOUT = 19
FLASH_ABORT_RECIEVED = 20
FLASH_PAGE_MISMATCH = 21
FLASH_BLOCK_MISMATCH = 22
FLASH_PAGE_NOT_IN_CODE_AREA = 23
MEM_ILLEGAL_ADDRESS = 24
FLASH_LOCKED = 25
INVALID_BLOCK = 26
FLASH_ILLEGAL_PAGE = 27
FLASH_TOO_MANY_BYTES = 28
FLASH_INVALID_STRING_NUM = 29

SMBUS_INQ_OVERFLOW = 32
SMBUS_OUTQ_UNDERFLOW = 33
SMBUS_CRC_FAILED = 34

SHT1x_COMM_TIME_OUT = 40
SHT1x_NO_ACK = 41
SHT1x_CRC_FAILED = 42
SHT1X_TOO_MANY_W_BYTES = 43
SHT1X_TOO_MANY_R_BYTES = 44
SHT1X_INVALID_MODE = 45
SHT1X_INVALID_LINE = 46

STREAM_IS_ACTIVE = 48
STREAM_TABLE_INVALID = 49
STREAM_CONFIG_INVALID = 50
STREAM_BAD_TRIGGER_SOURCE = 51
STREAM_NOT_RUNNING = 52
STREAM_INVALID_TRIGGER = 53
STREAM_ADC0_BUFFER_OVERFLOW = 54
STREAM_SCAN_OVERLAP = 55
STREAM_SAMPLE_NUM_INVALID = 56
STREAM_BIPOLAR_GAIN_INVALID = 57
STREAM_SCAN_RATE_INVALID = 58
STREAM_AUTORECOVER_ACTIVE = 59
STREAM_AUTORECOVER_REPORT = 60
STREAM_SOFTPWM_ON = 61
STREAM_INVALID_RESOLUTION = 63

PCA_INVALID_MODE = 64
PCA_QUADRATURE_AB_ERROR = 65
PCA_QUAD_PULSE_SEQUENCE = 66
PCA_BAD_CLOCK_SOURCE = 67
PCA_STREAM_ACTIVE = 68
PCA_PWMSTOP_MODULE_ERROR = 69
PCA_SEQUENCE_ERROR = 70
PCA_LINE_SEQUENCE_ERROR = 71
TMR_SHARING_ERROR = 72

EXT_OSC_NOT_STABLE = 80
INVALID_POWER_SETTING = 81
PLL_NOT_LOCKED = 82

INVALID_PIN = 96
PIN_CONFIGURED_FOR_ANALOG = 97
PIN_CONFIGURED_FOR_DIGITAL = 98
IOTYPE_SYNCH_ERROR = 99
INVALID_OFFSET = 100
IOTYPE_NOT_VALID = 101
INVALID_CODE = 102

UART_TIMEOUT = 112
UART_NOTCONNECTED = 113
UART_NOTENALBED = 114
I2C_BUS_BUSY = 116
TOO_MANY_BYTES = 118
TOO_FEW_BYTES = 119
DSP_PERIOD_DETECTION_ERROR = 128
DSP_SIGNAL_OUT_OF_RANGE = 129
MODBUS_RSP_OVERFLOW = 144
MODBUS_CMD_OVERFLOW = 145

class FIOPin(object):
    "Digital pin interface using the interface of nrf5x.py and drivers.py"
    def __init__(self, parent, number):
        self.parent = parent
        self.number = number

        io = parent.configIO()
        analogmask = io['FIOAnalog'] | (io['EIOAnalog'] << 8)
        # e.g., {'NumberOfTimersEnabled': 0, 'TimerCounterPinOffset': 4,
        #        'DAC1Enable': 0, 'FIOAnalog': 31, 'EIOAnalog': 0,
        #        'TimerCounterConfig': 64, 'EnableCounter1': False,
        #        'EnableCounter0': False}
        self.analog = bool(analogmask & (1<<number))

    def set(self):
        self._digital()
        self.parent.setFIOState(self.number, state=True)
    def clear(self):
        self._digital()
        self.parent.setFIOState(self.number, state=False)
    def read(self):
        self._digital()
        return bool(self.parent.getDIState(self.number))

    def _digital(self):
        if self.analog:
            self.parent.configDigital(self.number)
        self.analog = False

    @property
    def v(self):
        if self.number >= 16:
            raise NotImplemented("higher IO are digital-only")
        if not self.analog:
            self.analog = True
            self.parent.configAnalog(self.number)
        return self.parent.getAIN(self.number)

    @property
    def hiz(self):
        raise NotImplemented()

class DACPin(object):
    def __init__(self, parent, number):
        self.parent = parent
        self.number = number
        self._counts = None

    @property
    def v(self):
        return self.parent.dacBitsToVoltage(self.counts,
                                            dacNumber = self.number,
                                            is16Bits = True)

    @v.setter
    def v(self, volts):
        self.counts = self.parent.voltageToDACBits(volts,
                                                   dacNumber = self.number,
                                                   is16Bits = True)
    @property
    def counts(self):
        return self._counts

    @counts.setter
    def counts(self, c):
        c = int(round(c))

        c = max(0, c)
        c = min(0xFFFF, c)

        self._counts = c

        self.parent.getFeedback(u3py.DAC16(Dac = self.number,
                                           Value = self._counts))
class PWM(object):
    """LabJack PWM.
    Assumes it is the only timer / counter in use."""

    def __init__(self, pin):
        self.pin = pin
        self.parent = pin.parent
        self.parent.configTimerClock(TimerClockBase = 2)
        self.duty = 0.

    @property
    def duty(self):
        return (65536 - self._duty)/65536.

    @duty.setter
    def duty(self, v):
        self._duty = int(65536*(1.0-float(v)))
        if self._duty < 0:
            self._duty = 0
        if self._duty > 65536:
            self._duty = 65536
        if self._duty < 65536:
            if not self._pwming:
                self.parent.configIO(NumberOfTimersEnabled = 1,
                                     TimerCounterPinOffset = self.pin.number)
                self._pwming = True
            self.parent.getFeedback(u3py.Timer0Config(TimerMode = 0, Value = self._duty))
        else:
            self.parent.configIO(NumberOfTimersEnabled = 0)
            self.pin.clear()
            self._pwming = False

class UART(object):
    """Asynchronous receiver / transmitter using (a subset of) the interface
    of pySerial"""
    def __init__(self, txd_pin, rxd_pin=None, baudrate=None, timeout=None):

        if rxd_pin:
            assert txd_pin.parent == rxd_pin.parent # both pins must be same labjack
            assert txd_pin.number+1 == rxd_pin.number # must be order TXD, RXD.

        self.parent = txd_pin.parent

        self.parent.configIO(EnableUART = True,
                             TimerCounterPinOffset = txd_pin.number)

        self.baudrate = baudrate
        self.timeout = timeout
        self.readbuffer = b''

    @property
    def baudrate(self):
        ret = self.parent.asynchConfig(Update = False)
        def baud_from_factor(factor):
            # BaudFactor = (2**16) - 48000000//(2 * DesiredBaud)
            # BaudFactor - (2**16) = -48000000//(2 * DesiredBaud)
            # DesiredBaud * (BaudFactor - (2**16)) = -48000000//2
            # DesiredBaud = -48000000//(2 * BaudFactor - (2**16))
            return 48000000//(2 * (2**16 - factor))

        return baud_from_factor(ret['BaudFactor'])

    @baudrate.setter
    def baudrate(self, baud):
        ret = self.parent.asynchConfig(Update = True,
                                       UARTEnable = True,
                                       DesiredBaud = baud)

    def write(self, b):
        b = list(bytes(b))
        while b:
            this_b = b[:32]
            b = b[len(this_b):]
            ret = self.parent.asynchTX(this_b)
            assert ret['NumAsynchBytesSent'] == len(this_b)

            if ret['NumAsynchBytesInRXBuffer']:
                self._fill_rx_buffer()

    def read(self, size=1):
        import time
        t0 = time.time()
        while 1:
            if len(self.readbuffer) >= size:
                ret = self.readbuffer[:size]
                self.readbuffer = self.readbuffer[size:]
                return ret
            elif self.timeout is not None and time.time() > t0 + self.timeout:
                ret = self.readbuffer
                self.readbuffer = b''
                return ret

            self._fill_rx_buffer()

    def _fill_rx_buffer(self):
        rb = self.parent.asynchRX()
        ct = rb['NumAsynchBytesInRXBuffer']
        self.readbuffer += bytes(rb['AsynchBytes'][:ct])

class I2CAddressed(object):
    def __init__(self, parent, address):
        self.parent = parent
        self.address = address

    def write(self, command, data, stop=True):
        return self(I2CBytes = [command]+data,
                    NumI2CBytesToReceive = 0)

    def read(self, count, stop=True):
        return self(I2CBytes = [],
                    NumI2CBytesToReceive = count)

    def __call__(self, *args, **kwargs):
        kwargs.setdefault('Address', self.address)
        return self.parent(*args, **kwargs)

    def compat(self):
        return I2CAddressedCompat(self)

class I2CAddressedCompat(object):
    def __init__(self, parent):
        self.parent = parent

    def write(self, data):
        print('data',data)
        return self.exchange(data=data, count=0)

    def read(self, count):
        return self.exchange(data=[], count=count)

    def exchange(self, data, count):
        ret = self.parent(I2CBytes = list(data),
                          NumI2CBytesToReceive = count)
        print(ret)
        return bytes(ret['I2CBytes'])

class I2C(object):
    "LabJack I2C bus using the interface of nrf5x.py and drivers.py"
    def __init__(self, scl_pin, sda_pin):
        assert scl_pin.parent == sda_pin.parent # both pins must be same labjack
        self.parent = scl_pin.parent
        self.sda = sda_pin.number
        self.scl = scl_pin.number

    def address(self, address):
        return I2CAddressed(self, address)

    def __call__(self, *args, **kwargs):
        """ Here's a giant back door.

        Calling the object is the same as calling u3 objects's "i2c" method.

        So this is not at all the same interface as the nrf5x.py i2c objects.

        But *this* interface could be grafted on to nrf5x.py fairly easily.
        """

        kwargs.setdefault('I2CBytes',[])
        kwargs.setdefault('EnableClockStretching',True)
        kwargs.setdefault('NumI2CBytesToReceive', 0)
        #kwargs.setdefault('SpeedAdjust',200)
        #kwargs['ResetAtStart'] = True

        kwargs['SDAPinNum']=self.sda
        kwargs['SCLPinNum']=self.scl

        try:
            ret = self.parent.i2c(*args, **kwargs)
        except u3py.LowlevelErrorException as e:
            if e.errorCode==I2C_BUS_BUSY:
                raise ClockStuckException()
            else:
                raise e

        ackarray = bytes(ret['AckArray']+[0]*4)[:4]

        acked, = struct.unpack('<I', ackarray)
        bytelen = len(kwargs['I2CBytes'])

        if bytelen > 0 and bytelen <= 31:
            data = [(acked >> (b+1)) & 1 for b in range(bytelen)]
            message = {'Address':bytelen & 1,
                       'Data':[(bytelen >> (b+1)) & 1 for b in range(bytelen)]}

            # At least the address should be acknowledged.
            if 0==(acked & 1):
                raise NackException(message)

        return ret

class ManagedU3(u3py.U3):
    def __init__(self, *args, **kwargs):
        super(ManagedU3, self).__init__(*args, **kwargs)
        self.getCalibrationData()

        # assert self.deviceName in ['U3-HV']

        self.FIO = [FIOPin(self, pin) for pin in range(8)]
        self.EIO = [FIOPin(self, pin+8) for pin in range(8)]
        self.CIO = [FIOPin(self, pin+16) for pin in range(8)]

        for pin in range(8):
            setattr(self, 'FIO%d'%pin, self.FIO[pin])
        for pin in range(8):
            setattr(self, 'EIO%d'%pin, self.EIO[pin])
        for pin in range(8):
            setattr(self, 'CIO%d'%pin, self.CIO[pin])

        if self.deviceName == 'U3-HV':
            self.AIN = [self.FIO[pin] for pin in range(4)]
            for pin in range(4):
                setattr(self, 'AIN%d'%pin, self.AIN[pin])

        elif self.deviceName == 'U3-LV':
            self.AIN = [(self.FIO + self.EIO)[pin] for pin in range(8 + 8)]
            for pin in range(8 + 8):
                setattr(self, 'AIN%d'%pin, self.AIN[pin])

        else:
            raise Exception('Unknown LabJack type "%s"'%self.deviceName)

        self.DAC = [DACPin(self, pin) for pin in range(2)]
        for pin in range(2):
            setattr(self, 'DAC%d'%pin, self.DAC[pin])

    def dacBitsToVoltage(self, bits, dacNumber=0, is16Bits = False):
        """ Filling in a missing function in u3 API. This is just algrebra on voltageToDACBits() """

        if is16Bits:
            bits /= 256.

        if self.calData is not None:
            volts = (bits - self.calData['dac%sOffset' % dacNumber]) / self.calData['dac%sSlope' % dacNumber]
        else:
            volts = bits / 51.717

        return volts

    def i2c_bus(self, sda, scl):
        return I2C(sda_pin=sda, scl_pin=scl)

u3_devices = {}

def u3(*args, **kwargs):
    global u3_devices

    if ('serial' not in kwargs) or (kwargs['serial'] not in u3_devices):
        u3 = ManagedU3(*args, **kwargs)
        u3_devices[u3.serialNumber]=u3
        return u3

    return u3_devices[serial_number]

def unit_tests(lj):
    "connect FIO4 and FIO5 for these tests"

    lj.FIO5.set()
    assert lj.FIO4.read()
    lj.FIO5.clear()
    assert not lj.FIO4.read()

    lj.FIO5.set()
    assert lj.FIO4.read()
    lj.FIO5.clear()
    assert not lj.FIO4.read()

    lj.FIO5.set()
    assert lj.FIO4.v > 2.3
    lj.FIO5.clear()
    assert lj.FIO4.v < 2.5

    "connect DAC0 and AIN1 for these tests"
    def settle():
        import time
        time.sleep(0.1)

    lj.DAC0.v = 0.2
    settle()
    assert abs(lj.DAC0.v - lj.AIN1.v) < 0.1

    lj.DAC0.v = 4.2
    settle()
    assert abs(lj.DAC0.v - lj.AIN1.v) < 0.1

    lj.DAC0.v = 2.2
    settle()
    assert abs(lj.DAC0.v - lj.AIN1.v) < 0.1

    print("ok.")

if __name__=="__main__":
    #lj = u3(LJSocket="localhost:6000", serial=320042174)
    lj = u3()
    print(lj)
    print(lj.serialNumber)
    unit_tests(lj)



#!/usr/bin/env python3

import intelhex
import time
from .u3_manager import u3_manager
import itertools
from u3 import BitStateWrite, BitStateRead, WaitShort

"""

Program PIC microprocessors.

I hooked up like so:

| signal  | pin | color  | labjack |
|---------+-----+--------+---------|
| !MCLR   |   1 | brown  | FIO4    |
| +3V3    |   2 | red    | VS      |
| GND     |   3 | orange | GND     |
| ICSPDAT |   4 | yellow | FIO5    |
| ICSPCLK |   5 | green  | FIO6    |
| PGM     |   6 | blue   | --      |

"""

# this spaces out the timing so transactions are recognizable on the
# logic analyzer.  Speed penalty about ~2x 
dbg_flush = False

def dummy_print(*args, **kwargs): pass

# debug printing function, set to "print" for debug.
dbg_print = dummy_print

def reverse_bits(bit_count, val):
    ret = 0
    for x in range(bit_count):
        ret <<= 1
        ret |= val & 1
        val >>= 1
    return ret

class Pic:
    USERID_ADDR = 0x8000
    REVID_ADDR = 0x8005
    DEVID_ADDR = 0x8006

    def __init__(self,
                 mclr_pin=None,
                 icspdat_pin=None,
                 icspclk_pin=None,
                 skip_mclr = False):
        self.mclr = mclr_pin
        self.dat = icspdat_pin
        self.clk = icspclk_pin

        if not skip_mclr:
            self.mclr.set()
        self.dat.clear()
        self.clk.clear()

        self.tx_bits = 0
        self.tx_bits_ct = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not exc_type:
            self.flush_tx()

    @property
    def name(self):
        return self.__class__.__name__

    def send_value_msb(self, width, val):
        """ prepares to clock out value msb-first 

        val is value to send (integer)
        width is number of bits to send
        """        
        return self.send_value_lsb(width, reverse_bits(width, val))
        
    def send_value_lsb(self, width, val):
        """ prepares to clock out value lsb-first 

        val is value to send (integer)
        width is number of bits to send
        """
        dbg_print(f"sending {width} bits {val:x}")
        
        assert val >= 0
        mask = (1 << width) - 1
        assert 0 == val & (~mask)
        
        self.tx_bits |= val << self.tx_bits_ct
        self.tx_bits_ct += width
                
    def _flush_tx_bits(self):

        lj = self.clk.parent
        fb_commands = []

        # I measure 9 us between command executions
        # setup and hold times are 100 ns, so no need to add delays
        
        dbg_print(f"bit flushing {self.tx_bits_ct} bits {self.tx_bits:x}")
        for bit_num in range(self.tx_bits_ct):
            bit = self.tx_bits & 1
            fb_commands += [                
                BitStateWrite(self.dat.number, bit),
                BitStateWrite(self.clk.number, 1),
                # WaitShort(Time = setup_time/128e-6)
                BitStateWrite(self.clk.number, 0)
                # WaitShort(Time = hold_time/128e-6)
            ]
            self.tx_bits >>= 1
        self.tx_bits_ct = 0
            
        lj.getFeedback(fb_commands)

    def _flush_tx_spi(self):

        dbg_print(f"spi flushing {self.tx_bits_ct} bits {self.tx_bits:x}")
        
        spi_bytes = []
        while self.tx_bits_ct >= 8:
            spi_bytes.append(reverse_bits(8, self.tx_bits & 0xff))
            self.tx_bits >>= 8
            self.tx_bits_ct -= 8
    
        unused = 0xff
        lj = self.clk.parent

        chunklen = 50

        for pos in range(0, len(spi_bytes), chunklen):
            lj.spi(SPIBytes = spi_bytes[pos:pos+chunklen],
                   CLKPinNum = self.clk.number,
                   MOSIPinNum = self.dat.number,
                   MISOPinNum = unused,
                   CSPinNum = unused,
                   SPIMode='B',
                   AutoCS=False)

    def flush_tx(self):
        self._flush_tx_spi()
        self._flush_tx_bits()
            
    def read_value_lsb(self, width):
        """ width is number of bytes to read (lsb-first)
        return value is integer
        """        
        ret = self.read_value_msb(width)
        return reverse_bits(8*width, ret)

    def read_value_msb(self, width):
        """ width is number of bytes to read (msb-first)
        return value is integer
        """
        self.flush_tx()

        unused = 0xff
        lj = self.clk.parent
        spi = lj.spi(SPIBytes = [0]*width,
                     CLKPinNum = self.clk.number,
                     MOSIPinNum = unused,
                     MISOPinNum = self.dat.number,
                     CSPinNum = unused,
                     SPIMode='B',
                     AutoCS=False)
        ret = 0
        for i in range(width):
            ret <<= 8
            ret += spi['SPIBytes'][i]

        dbg_print("read %x"%ret)
        return ret

    def sleep(self, timeout):
        if timeout > 4e-6: # timeouts less than this are taken-care of
                           # by the slow SPI clock
            self.flush_tx()
            time.sleep(timeout)

class Pic16F_Enhanced_Midrange(Pic):

    USER_ADDR = 0x8000

    CONFIG_ADDR = 0x8007
    CONFIG_LEN = 0xC
    DIA_ADDR = 0x8100
    DIA_LEN = 0x40
    EEPROM_ADDR = 0xF000
    EEPROM_LEN = 0x100

    CONFIG_LEN = 5
    
    KEY = 0x4d434850 # 'MCHP'

    Tenth = 250e-6
    Tents = 100e-9
    Tckh = 100e-9
    Tckl = 100e-9
    Tdly = 1e-6
    Tpint = 0.002 # program memory
    Tpint_config = 0.0056 # config memory

    def send8_24(self, d1, d2):
        self.send8(d1)
        self.sleep(self.Tdly)
        self.send24(d2<<1)
        self.sleep(self.Tdly)

    def send32(self, d): self.send_value_msb(32, d)
    def send16(self, d): self.send_value_msb(16, d)
    def send24(self, d): self.send_value_msb(24, d)
    def send8(self, d): self.send_value_msb(8, d)

    def read24(self):
        start_bit = (1<<31)
        stop_bit = (1<<0)
        bits_mask = ~(start_bit | stop_bit)

        return (self.read_value_msb(24//8) & bits_mask) >> 1

    def enter_lvp(self):
        self.mclr.clear()
        self.sleep(self.Tenth)
        self.send32(self.KEY)
        self.sleep(self.Tenth)

    def set_PC(self, pc):
        self.pc = pc
        self.send8_24(0x80, pc)

    def increment_pc(self):
        self.pc += 1
        self.send8(0xf8)

    def load_mem(self, d, inc=False):
        "loads one word of data from PC into buffer on target"
        self.send8_24([0x00, 0x02][inc], d)
        self.pc += inc

    def write_row(self):
        self.send8(0xe0) # internally timed
        if self.pc < self.USER_ADDR:
            self.sleep(self.Tpint)
        else:
            self.sleep(self.Tpint_config)

    def read_mem(self, inc=True):
        if inc:
            self.send8(0xfe)
        else:
            self.send8(0xfc)

        self.sleep(self.Tdly)
        ret = self.read24()
        self.pc += inc
        return ret

    def read_word(self, address):
        self.set_PC(address)
        return self.read_mem()

    def read_device_id(self):
        return self.read_word(self.DEVID_ADDR)

    def read_config_stuff(self):
        mmap = (("user id 0", 0x8000),
                ("user id 1", 0x8001),
                ("user id 2", 0x8002),
                ("user id 3", 0x8003),
                ("reserved", 0x8004),
                ("revision id", 0x8005),
                ("device id", 0x8006),
                ("config 0", 0x8007),
                ("config 1", 0x8008),
                ("config 2", 0x8009),
                ("config 3", 0x800a),
                ("config 4", 0x800b))
        for name,addr in mmap:
            dbg_print(name, "@%04x:"%addr, "%04x"%self.read_word(addr))

    def exit_lvp(self):
        self.flush_tx()
        self.mclr.set()

    def bulk_erase(self,
                   erase_eeprom = False,
                   erase_flash = False,
                   erase_user_id = False,
                   erase_config = False):
        cmd = sum([erase_eeprom << 0,
                   erase_flash << 1,
                   erase_user_id << 2,
                   erase_config << 3])

        #dbg_print("Erase cmd:",cmd)

        self.send8_24(0x18, cmd)
        self.sleep(self.Terab)

    @property
    def Terab(self):
        if self.FLASH_LEN <= 8192:
            return 0.010
        else:
            return 0.013

    def read_flash(self, start_addr = None, length = None):
        if start_addr is None:
            start_addr = 0
        if length is None:
            length = self.FLASH_LEN

        self.set_PC(start_addr)
        return [self.read_mem() for _ in range(length)]

    def write_flash(self, data, start_addr = None):
        if start_addr is None:
            start_addr = 0

        # This is the best way to write rows.
        # Any partial-row algo risks writing garbage left over in row latch

	# See note on p19 of spec about write sizes for different algorithms.
        if start_addr < self.FLASH_LEN:
            row_size = 32
        else:
            row_size = 1

        front_pad = start_addr % row_size
        start_addr -= front_pad
        back_pad = row_size - 1 - ((len(data)+row_size-1) % row_size)

        # dbg_print("data",data)
        # dbg_print("pads",front_pad, back_pad)

        data = [0x3fff] * front_pad + data + [0x3fff] * back_pad

        self.set_PC(start_addr)

        while data:
            row = data[:row_size]
            data = data[row_size:]

            for d in row[:(row_size-1)]:
                self.load_mem(d, inc=True)

            self.load_mem(row[row_size-1], inc=False)
            self.write_row()

            #dbg_print("write row @pc=%04x"%self.pc)
            self.increment_pc()

    @property
    def memory_map(self):
        """ Returns list of tuples:
        (start, stop, name, erase block)

        name is a string, one of "program", "config", "user", "eeprom"

        """

        return [
            (0, self.FLASH_LEN, "program", 32),
            (0x8000, 0x8004, "user", 4),
            (0x8007, 0x8007 + self.CONFIG_LEN, "config", 4),
            (0xf000, 0xf100, "eeprom", 1),
        ]

class Pic16F_40002317(Pic16F_Enhanced_Midrange):
    """
    Programming spec: 40002317

    https://ww1.microchip.com/downloads/aemDocuments/documents/MCU08/ProductDocuments/ProgrammingSpecifications/PIC16F180XX-Family-Programming-Specification-40002317.pdf
    """

class Pic16F18013(Pic16F_40002317):
    FLASH_LEN = 2*1024 # words
class Pic16F18014(Pic16F_40002317):
    FLASH_LEN = 4*1024
class Pic16F18015(Pic16F_40002317):
    FLASH_LEN = 8*1024
class Pic16F18023(Pic16F_40002317):
    FLASH_LEN = 2*1024
class Pic16F18024(Pic16F_40002317):
    FLASH_LEN = 4*1024
class Pic16F18025(Pic16F_40002317):
    FLASH_LEN = 8*1024
class Pic16F18026(Pic16F_40002317):
    FLASH_LEN = 16*1024
class Pic16F18044(Pic16F_40002317):
    FLASH_LEN = 4*1024
class Pic16F18045(Pic16F_40002317):
    FLASH_LEN = 8*1024
class Pic16F18046(Pic16F_40002317):
    FLASH_LEN = 16*1024
class Pic16F18054(Pic16F_40002317):
    FLASH_LEN = 4*1024
class Pic16F18055(Pic16F_40002317):
    FLASH_LEN = 8*1024
class Pic16F18056(Pic16F_40002317):
    FLASH_LEN = 16*1024
class Pic16F18074(Pic16F_40002317):
    FLASH_LEN = 4*1024
class Pic16F18075(Pic16F_40002317):
    FLASH_LEN = 8*1024
class Pic16F18076(Pic16F_40002317):
    FLASH_LEN = 16*1024

class Pic16F_40002266(Pic16F_Enhanced_Midrange):
    """
    Programming spec: 40002266

    https://ww1.microchip.com/downloads/aemtest/MCU08/ProductDocuments/ProgrammingSpecifications/PIC16F171XX-Family-Programming-Specification-40002266.pdf
    """
    Tpint = 0.0028 # program memory

class Pic16F17114(Pic16F_40002266):
    FLASH_LEN = 4*1024
class Pic16F17115(Pic16F_40002266):
    FLASH_LEN = 8*1024
class Pic16F17124(Pic16F_40002266):
    FLASH_LEN = 4*1024
class Pic16F17125(Pic16F_40002266):
    FLASH_LEN = 8*1024
class Pic16F17126(Pic16F_40002266):
    FLASH_LEN = 16*1024
class Pic16F17144(Pic16F_40002266):
    FLASH_LEN = 4*1024
class Pic16F17145(Pic16F_40002266):
    FLASH_LEN = 8*1024
class Pic16F17146(Pic16F_40002266):
    FLASH_LEN = 16*1024
class Pic16F17154(Pic16F_40002266):
    FLASH_LEN = 4*1024
class Pic16F17155(Pic16F_40002266):
    FLASH_LEN = 8*1024
class Pic16F17156(Pic16F_40002266):
    FLASH_LEN = 16*1024
class Pic16F17174(Pic16F_40002266):
    FLASH_LEN = 4*1024
class Pic16F17175(Pic16F_40002266):
    FLASH_LEN = 8*1024
class Pic16F17176(Pic16F_40002266):
    FLASH_LEN = 16*1024

class Pic16F_XLP(Pic16F_Enhanced_Midrange):
    # these from Table 4-1
    COMMAND_LOAD_CONFIG = 0x00
    COMMAND_LOAD_DATA = 0x02
    COMMAND_READ_DATA = 0x04
    COMMAND_INC_ADDRESS = 0x06
    COMMAND_RESET_ADDRESS = 0x16
    COMMAND_BEGIN_PROG_INT = 0x08
    COMMAND_BEGIN_PROG_EXT = 0x18
    COMMAND_END_PROG_EXT = 0x0a
    COMMAND_BULK_ERASE = 0x09
    COMMAND_ROW_ERASE = 0x11

    CONFIG_LEN = 2
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pc = 99999999

    def increment_pc(self):
        self.send6_le(self.COMMAND_INC_ADDRESS) 
        self.pc += 1

    def set_PC(self, address):
        
        dbg_print("setting PC=%04x to %04x"%(self.pc, address))
        if address >= 0x8000: # config data
            if (self.pc < 0x8000) or (self.pc > address):
                dbg_print("  loading config")
                self.send6_le(self.COMMAND_LOAD_CONFIG)
                self.send16_le(0)
                self.pc = 0x8000
        else: # not config data
            if self.pc > address:
                dbg_print("  resetting address")
                self.send6_le(self.COMMAND_RESET_ADDRESS)
                self.pc = 0

        while self.pc < address:
            self.increment_pc()

    def read_mem(self, inc=True):
        self.send6_le(self.COMMAND_READ_DATA)
        self.sleep(self.Tdly)
        ret = self.read16_le()

        if inc:
            self.increment_pc()

        return ret

    def read16_le(self):
        "fourteen bits packed into sixteen clocks"
        start_bit = (1<<15)
        stop_bit = (1<<0)
        bits_mask = ~(start_bit | stop_bit)

        ret = (self.read_value_lsb(16//8) & bits_mask) >> 1
        return ret

    def send32_le(self, d):
        self.send_value_lsb(32, d)

    def send16_le(self, d):
        "fourteen bits packed into sixteen clocks"
        return self.send_value_lsb(16, (d & 0x3fff) << 1)

    def send6_le(self, cmd):
        ret = self.send_value_lsb(6, cmd)
        if dbg_flush: self.flush_tx()
        return ret
    
    def enter_lvp(self):
        self.mclr.clear()
        self.sleep(self.Tenth)
        self.send32_le(self.KEY)
        self.send_value_lsb(1, 0)  # 33 bits!
        self.sleep(self.Tenth)

    def load_prog_mem(self, d, inc):
        dbg_print("load_prog %04x %03x"%(self.pc, d))
        self.send6_le(self.COMMAND_LOAD_DATA)        
        self.send16_le(d)
        if inc:
            self.increment_pc()

    def load_config_mem(self, d, inc):
        addr = self.pc
        
        dbg_print("load_config %04x %03x"%(self.pc, d))
        self.send6_le(self.COMMAND_LOAD_CONFIG)
        self.send16_le(d)
        self.pc = 0x8000
        while self.pc < addr:
            self.increment_pc()
            
        if inc:
            self.increment_pc()
            
    def load_mem(self, d, inc=False):
        "loads one word of data from PC into buffer on target"
        if self.pc >= 0x8000:
            return self.load_config_mem(d, inc)
        else:
            return self.load_prog_mem(d, inc)

    def write_row_prog_int(self):
        self.send6_le(self.COMMAND_BEGIN_PROG_INT)
        self.sleep(self.Tpint)

    def write_row_prog_ext(self):
        """not measurably faster than internal
        (due to USB timing no doubt.)"""
        self.send6_le(self.COMMAND_BEGIN_PROG_EXT)
        self.sleep(self.Tpext_min)
        self.send6_le(self.COMMAND_END_PROG_EXT)
        self.sleep(self.Tdis)      
                
    def write_row_config(self):
        self.send6_le(self.COMMAND_BEGIN_PROG_INT)
        self.sleep(self.Tpint_config)
            
    def write_row(self):
        " Write a row of data in one shot "
        if self.pc >= 0x8000:
            return self.write_row_config()
        else:
            return self.write_row_prog_int()
        
    def bulk_erase(self,
                   erase_eeprom = False,
                   erase_flash = False,
                   erase_user_id = False,
                   erase_config = False):

        if (erase_flash and
            (not erase_user_id) and
            erase_config):
            self.set_PC(0)
            self.send6_le(self.COMMAND_BULK_ERASE)

        elif (erase_flash and
              erase_user_id and
              erase_config):
            self.set_PC(0x8000)
            self.send6_le(self.COMMAND_BULK_ERASE)

        else:
            raise Exception("This bulk erase is not possible")
                    
        self.sleep(self.Terab)


class Pic16F_40001683(Pic16F_XLP):
    """
    Programming spec: 40001683

    https://ww1.microchip.com/downloads/en/DeviceDoc/40001683B.pdf
    """

    Tpint = 0.0025 # program memory
    Tpint_config = 0.005 # config memory
    Tpext_min = 0.001
    Tdis = 0.0003
    
class Pic16F1703(Pic16F_40001683):
    FLASH_LEN = 0x0800
class Pic16F1704(Pic16F_40001683):
    FLASH_LEN = 0x1000
class Pic16F1705(Pic16F_40001683):
    FLASH_LEN = 0x2000
class Pic16F1707(Pic16F_40001683):
    FLASH_LEN = 0x0800
class Pic16F1708(Pic16F_40001683):
    FLASH_LEN = 0x1000
class Pic16F1709(Pic16F_40001683):
    FLASH_LEN = 0x2000

class Pic16LF1703(Pic16F1703): pass
class Pic16LF1704(Pic16F1704): pass
class Pic16LF1705(Pic16F1705): pass
class Pic16LF1707(Pic16F1707): pass
class Pic16LF1708(Pic16F1708): pass
class Pic16LF1709(Pic16F1709): pass

# DS40001683B-page 9

device_ids = {
    0x30DB: Pic16F17114,
    0x30E2: Pic16F17115,
    0x30DC: Pic16F17124,
    0x30DE: Pic16F17125,
    0x30E0: Pic16F17126,
    0x30DD: Pic16F17144,
    0x30DF: Pic16F17145,
    0x30E1: Pic16F17146,
    0x3101: Pic16F17154,
    0x3103: Pic16F17155,
    0x3105: Pic16F17156,
    0x3102: Pic16F17174,
    0x3104: Pic16F17175,
    0x3106: Pic16F17176,

    0x30F1: Pic16F18013,
    0x30F2: Pic16F18014,
    0x30F5: Pic16F18015,
    0x30F3: Pic16F18023,
    0x30F4: Pic16F18024,
    0x30F6: Pic16F18025,
    0x30F9: Pic16F18026,
    0x30F7: Pic16F18044,
    0x30F8: Pic16F18045,
    0x30FA: Pic16F18046,
    0x30FB: Pic16F18054,
    0x30FC: Pic16F18055,
    0x30FF: Pic16F18056,
    0x30FD: Pic16F18074,
    0x30FE: Pic16F18075,
    0x3100: Pic16F18076,

    0x3061: Pic16F1703,
    0x3063: Pic16LF1703,
    0x3043: Pic16F1704,
    0x3045: Pic16LF1704,
    0x3055: Pic16F1705,
    0x3057: Pic16LF1705,
    0x3060: Pic16F1707,
    0x3062: Pic16LF1707,
    0x3042: Pic16F1708,
    0x3044: Pic16LF1708,
    0x3054: Pic16F1709,
    0x3056: Pic16LF1709,
}

def specialize(general_pic, cls):
    return cls(mclr_pin = general_pic.mclr,
               icspdat_pin = general_pic.dat,
               icspclk_pin = general_pic.clk,
               skip_mclr = True)

def program(mclr_pin,
            icspdat_pin,
            icspclk_pin,
            hex_filename,
            require_pic=None):

    found = False
    found_ids = []

    # This dance is required because different classes of PIC have
    # different readout mechanisms
    for cls in [Pic16F_Enhanced_Midrange,
                Pic16F_XLP]:
        general_pic = cls(mclr_pin = mclr_pin,
                          icspdat_pin = icspdat_pin,
                          icspclk_pin = icspclk_pin)
        
        general_pic.enter_lvp()
        device_id = general_pic.read_device_id()

        found_ids.append(device_id)
        
        if device_id in device_ids:
            found = True
            break
        general_pic.exit_lvp()
                    
    if not found:
        raise Exception("No known PIC found:",
              ", ".join('0x%04x'%did for did in found_ids))
    
    with specialize(general_pic, device_ids[device_id]) as pic:

        print("Found", pic.name)
        if require_pic:
            assert require_pic.upper() == pic.name.upper()

        ih = intelhex.IntelHex16bit()
        ih.loadfile(hex_filename, format='hex')

        used_regions = []
        for seg_start, seg_stop in ih.segments():
            #dbg_print("seg", seg_start, seg_stop)
            # The spec is to have intel-hex addresses at twice the PIC address.
            assert 0 == seg_start % 2
            assert 0 == seg_stop % 2

            seg_start //= 2
            seg_stop //= 2

            assigned_regions = []

            for mem_start, mem_stop, name, erase_size in pic.memory_map:
                if mem_start >= seg_stop:
                    continue
                if seg_start >= mem_stop:
                    continue

                # round down to erase row
                seg_start = (seg_start//erase_size)*erase_size
                # round up to erase row size
                seg_stop  = ((seg_stop + erase_size - 1)//erase_size)*erase_size

                assigned_regions.append([name, seg_start, seg_stop])

            assert len(assigned_regions) <= 1
            used_regions.extend(assigned_regions)

            # Coalesce contiguous
            cont_regions = []
            last_region = None
            for region in used_regions:
                if last_region:
                    if (region[0] == last_region[0] and # compare names
                        last_region[2]>=region[1]): # contiguous
                        last_region[2]=region[2]
                    else:
                        cont_regions.append(last_region)
                        last_region = region
                else:
                    last_region = region
            if last_region:
                cont_regions.append(last_region)

        #dbg_print("used regions", used_regions)
        #dbg_print("contiguous regions", cont_regions)

        region_by_type = {}
        for c in cont_regions:
            region_by_type.setdefault(c[0], []).append(c[1:])

        print(region_by_type)
        pic.bulk_erase(#erase_eeprom = 'eeprom' in region_by_type,
                       erase_flash = 'program' in region_by_type,
                       erase_user_id = 'user' in region_by_type,
                       erase_config = 'config' in region_by_type)

        verify_errors = 0

        for r, reglist in region_by_type.items():
            for reg in reglist:
                words = [ih[x] for x in range(*reg)]
                #dbg_print(r, words, ["%04x"%x for x in reg])
                #dbg_print(">  ",pic.read_flash(start_addr = reg[0], length=reg[1]-reg[0]))
                pic.write_flash(words, start_addr = reg[0])
                #dbg_print(">> ",pic.read_flash(start_addr = reg[0], length=reg[1]-reg[0]))
                readback = pic.read_flash(start_addr = reg[0], length=reg[1]-reg[0])
                for word_wrote,word_read,addr in zip(words, readback, range(*reg)):
                    if (word_wrote ^ word_read) & 16383:
                        for start,end,name,_ in pic.memory_map:
                            if addr >= start and addr < end:
                                print(f"Verify Error at {addr:04x} (wrote {word_wrote:04x}, read {word_read:04x})")
                                verify_errors += 1

            #pic.read_config_stuff()

        pic.exit_lvp()
        return verify_errors
    
def main():
    import argparse
    parser = argparse.ArgumentParser(description="Program a PIC using LabJack.")
    parser.add_argument("--pic", "-p", help='Require this device before programming')
    parser.add_argument("hexfile", metavar="HEXFILE", nargs=1, help="Hex file to program")

    parser.add_argument(
        "--mclr", default='FIO4', help="Labjack pin for !MCLR. (Default %(default)s)"
    )

    parser.add_argument(
        "--icspdat", default='FIO5', help="Labjack pin for ICSPDAT. (Default %(default)s)"
    )

    parser.add_argument(
        "--icspclk", default='FIO6', help="Labjack pin for ICSPCLK. (Default %(default)s)"
    )

    args = parser.parse_args()

    lj = u3_manager.u3()
    return program(mclr_pin = getattr(lj, args.mclr),
                   icspdat_pin = getattr(lj, args.icspdat),
                   icspclk_pin = getattr(lj, args.icspclk),
                   require_pic = args.pic,
                   hex_filename = args.hexfile[0])

if __name__=="__main__":
    main()
